@@ -9,7 +9,9 @@ from PIL import Image
 from reversebox.common.common import calculate_padding_length
 from reversebox.common.logger import get_logger
 from reversebox.image.compression.compression_gst import decompress_gst_image
+from reversebox.image.decoders.yuv_decoder import YUVDecoder
 from reversebox.image.image_formats import ImageFormats
+from reversebox.image.pillow_wrapper import PillowWrapper
 from reversebox.image.swizzling.swizzle_gst import (
     unswizzle_detail1,
     unswizzle_detail2,
@@ -30,9 +32,20 @@ logger = get_logger(__name__)
 
 
 class ImageDecoder:
+
+    pillow_wrapper = PillowWrapper()
+
     def __init__(self):
         pass
 
+    def _decode_rgb121_byte_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (pixel_int >> 3) * 255
+        p[1] = ((pixel_int >> 1) & 3) * 85
+        p[2] = (pixel_int & 1) * 255
+        p[3] = 0xFF
+        return p
+
     def _decode_rgbx2222_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
         p[0] = (pixel_int & 3) * 85
@@ -49,11 +62,51 @@ class ImageDecoder:
         p[3] = ((pixel_int >> 6) & 3) * 85
         return p
 
+    def _decode_gray8_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = pixel_int & 255
+        p[1] = pixel_int & 255
+        p[2] = pixel_int & 255
+        p[3] = 0xFF
+        return p
+
+    def _decode_rgb332_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (pixel_int >> 5) * 36
+        p[1] = ((pixel_int >> 2) & 7) * 36
+        p[2] = (pixel_int & 3) * 85
+        p[3] = 0xFF
+        return p
+
+    def _decode_bgr332_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (pixel_int & 7) * 36
+        p[1] = ((pixel_int >> 3) & 7) * 36
+        p[2] = (pixel_int >> 6) * 85
+        p[3] = 0xFF
+        return p
+
     def _decode_rgbx5551_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
-        r = pixel_int & 0x1F
+        p[0] = (((pixel_int >> 11) & 0x1F) * 0xFF // 0x1F)
+        p[1] = (((pixel_int >> 6) & 0x1F) * 0xFF // 0x1F)
+        p[2] = (((pixel_int >> 1) & 0x1F) * 0xFF // 0x1F)
+        p[3] = 0xFF
+        return p
+
+    def _decode_rgba5551_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (((pixel_int >> 11) & 0x1F) * 0xFF // 0x1F)
+        p[1] = (((pixel_int >> 6) & 0x1F) * 0xFF // 0x1F)
+        p[2] = (((pixel_int >> 1) & 0x1F) * 0xFF // 0x1F)
+        p[3] = (((pixel_int >> 0) & 0x1F) * 0xFF // 0x1F)
+        return p
+
+    def _decode_bgrx5551_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        b = pixel_int & 0x1F
         g = (pixel_int >> 5) & 0x1F
-        b = (pixel_int >> 10) & 0x1F
+        r = (pixel_int >> 10) & 0x1F
         p[0] = (r << 3) | (r >> 2)
         p[1] = (g << 3) | (g >> 2)
         p[2] = (b << 3) | (b >> 2)
@@ -65,9 +118,9 @@ class ImageDecoder:
         r = pixel_int & 0x1F
         g = (pixel_int >> 5) & 0x1F
         b = (pixel_int >> 10) & 0x1F
-        p[2] = (r << 3) | (r >> 2)
-        p[1] = (g << 3) | (g >> 2)
         p[0] = (b << 3) | (b >> 2)
+        p[1] = (g << 3) | (g >> 2)
+        p[2] = (r << 3) | (r >> 2)
         p[3] = 0xFF
         return p
 
@@ -102,14 +155,22 @@ class ImageDecoder:
         p[3] = 0xFF
         return p
 
+    def _decode_bgr565_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = ((pixel_int >> 0) & 0x1F) * 0xFF // 0x1F
+        p[1] = ((pixel_int >> 5) & 0x3F) * 0xFF // 0x3F
+        p[2] = ((pixel_int >> 11) & 0x1F) * 0xFF // 0x1F
+        p[3] = 0xFF
+        return p
+
     def _decode_rgb5A3_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
-        if pixel_int & 0x8000 != 0:  # r5g5b5
+        if pixel_int & 0x8000:  # rgb555
             p[0] = (((pixel_int >> 10) & 0x1F) * 0xFF // 0x1F)
             p[1] = (((pixel_int >> 5) & 0x1F) * 0xFF // 0x1F)
             p[2] = (((pixel_int >> 0) & 0x1F) * 0xFF // 0x1F)
             p[3] = 0xFF
-        else:  # r4g4b4a3
+        else:  # argb3444
             p[0] = (((pixel_int >> 8) & 0x0F) * 0xFF // 0x0F)
             p[1] = (((pixel_int >> 4) & 0x0F) * 0xFF // 0x0F)
             p[2] = (((pixel_int >> 0) & 0x0F) * 0xFF // 0x0F)
@@ -142,26 +203,26 @@ class ImageDecoder:
 
     def _decode_bgra8888_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
-        p[2] = (pixel_int >> 0) & 0xff
-        p[1] = (pixel_int >> 8) & 0xff
         p[0] = (pixel_int >> 16) & 0xff
-        p[3] = (pixel_int >> 24) & 0xff
-        return p
-
-    def _decode_argb8888_pixel(self, pixel_int: int) -> bytes:
-        p = bytearray(4)
-        p[2] = (pixel_int >> 0) & 0xff
         p[1] = (pixel_int >> 8) & 0xff
-        p[0] = (pixel_int >> 16) & 0xff
+        p[2] = (pixel_int >> 0) & 0xff
         p[3] = (pixel_int >> 24) & 0xff
         return p
 
     def _decode_abgr8888_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
+        p[0] = (pixel_int >> 24) & 0xff
+        p[1] = (pixel_int >> 16) & 0xff
+        p[2] = (pixel_int >> 8) & 0xff
         p[3] = (pixel_int >> 0) & 0xff
+        return p
+
+    def _decode_argb8888_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
         p[0] = (pixel_int >> 8) & 0xff
         p[1] = (pixel_int >> 16) & 0xff
         p[2] = (pixel_int >> 24) & 0xff
+        p[3] = (pixel_int >> 0) & 0xff
         return p
 
     def _decode_argb4444_pixel(self, pixel_int: int) -> bytes:
@@ -190,60 +251,114 @@ class ImageDecoder:
         p[3] = (a << 4) | (a >> 0)
         return p
 
-    def _decode_yuy2_pixel(self, Y: float, U: float, V: float) -> bytes:
+    def _decode_rgbx4444_pixel(self, pixel_int: int) -> bytes:
         p = bytearray(4)
+        r = (pixel_int >> 8) & 0x0f
+        g = (pixel_int >> 4) & 0x0f
+        b = (pixel_int >> 0) & 0x0f
 
-        def _round_clamp_int(f: float) -> int:
-            i: int = int(f+0.5)
-            if i < 0:
-                i = 0
-            if i > 255:
-                i = 255
-            return i
-
-        C: float = Y - 16.0
-        D: float = U - 128.0
-        E: float = V - 128.0
-
-        R: float = 1.164383 * C + 1.596027 * E
-        G: float = 1.164383 * C - (0.391762 * D) - (0.812968 * E)
-        B: float = 1.164383 * C + 2.017232 * D
-
-        p[0] = _round_clamp_int(R)
-        p[1] = _round_clamp_int(G)
-        p[2] = _round_clamp_int(B)
+        p[0] = (b << 4) | (b >> 0)
+        p[1] = (g << 4) | (g >> 0)
+        p[2] = (r << 4) | (r >> 0)
         p[3] = 0xFF
         return p
 
+    def _decode_rgbx6666_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (((pixel_int >> 0) & 63) * 255 + 32) // 63
+        p[1] = (((pixel_int >> 8) & 63) * 255 + 32) // 63
+        p[2] = (((pixel_int >> 16) & 63) * 255 + 32) // 63
+        p[3] = 0xFF
+        return p
+
+    def _decode_i4_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = pixel_int * 0x11
+        p[1] = pixel_int * 0x11
+        p[2] = pixel_int * 0x11
+        p[3] = 0xFF
+        return p
+
+    def _decode_i8_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = pixel_int
+        p[1] = pixel_int
+        p[2] = pixel_int
+        p[3] = 0xFF
+        return p
+
+    def _decode_ia4_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = (pixel_int & 0xF) * 0x11
+        p[1] = (pixel_int & 0xF) * 0x11
+        p[2] = (pixel_int & 0xF) * 0x11
+        p[3] = (pixel_int >> 4) * 0x11
+        return p
+
+    def _decode_ia8_pixel(self, pixel_int: int) -> bytes:
+        p = bytearray(4)
+        p[0] = pixel_int & 0xFF
+        p[1] = pixel_int & 0xFF
+        p[2] = pixel_int & 0xFF
+        p[3] = pixel_int >> 8
+        return p
+
     generic_data_formats = {
         # image_format: (decode_function, bits_per_pixel, image_entry_read_function)
+        ImageFormats.RGB121: (_decode_rgb121_byte_pixel, 4, get_uint8),
+        ImageFormats.RGB121_BYTE: (_decode_rgb121_byte_pixel, 8, get_uint8),
+        ImageFormats.GRAY8: (_decode_gray8_pixel, 8, get_uint8),
         ImageFormats.RGBX2222: (_decode_rgbx2222_pixel, 8, get_uint8),
         ImageFormats.RGBA2222: (_decode_rgba2222_pixel, 8, get_uint8),
+        ImageFormats.RGB332: (_decode_rgb332_pixel, 8, get_uint8),
+        ImageFormats.BGR332: (_decode_bgr332_pixel, 8, get_uint8),
         ImageFormats.RGB565: (_decode_rgb565_pixel, 16, get_uint16),
+        ImageFormats.BGR565: (_decode_bgr565_pixel, 16, get_uint16),
+        ImageFormats.RGBX5551: (_decode_rgbx5551_pixel, 16, get_uint16),
+        ImageFormats.RGBA5551: (_decode_rgba5551_pixel, 16, get_uint16),
         ImageFormats.RGB888: (_decode_rgb888_pixel, 24, get_uint24),
         ImageFormats.BGR888: (_decode_bgr888_pixel, 24, get_uint24),
         ImageFormats.ARGB4444: (_decode_argb4444_pixel, 16, get_uint16),
         ImageFormats.RGBA4444: (_decode_rgba4444_pixel, 16, get_uint16),
+        ImageFormats.RGBX4444: (_decode_rgbx4444_pixel, 16, get_uint16),
         ImageFormats.XRGB1555: (_decode_xrgb1555_pixel, 16, get_uint16),
         ImageFormats.ABGR1555: (_decode_abgr1555_pixel, 16, get_uint16),
         ImageFormats.XBGR1555: (_decode_xbgr1555_pixel, 16, get_uint16),
         ImageFormats.ARGB8888: (_decode_argb8888_pixel, 32, get_uint32),
         ImageFormats.ABGR8888: (_decode_abgr8888_pixel, 32, get_uint32),
+        ImageFormats.RGBA8888: (_decode_rgba8888_pixel, 32, get_uint32),
+        ImageFormats.BGRA8888: (_decode_bgra8888_pixel, 32, get_uint32),
+        ImageFormats.N64_RGB5A3: (_decode_rgb5A3_pixel, 16, get_uint16),
+        ImageFormats.N64_I4: (_decode_i4_pixel, 4, get_uint8),
+        ImageFormats.N64_I8: (_decode_i8_pixel, 8, get_uint8),
+        ImageFormats.N64_IA4: (_decode_ia4_pixel, 8, get_uint8),
+        ImageFormats.N64_IA8: (_decode_ia8_pixel, 16, get_uint16),
     }
 
     indexed_data_formats = {
         # image_format: (decode_function, bits_per_pixel, palette_entry_size, palette_entry_read_function)
         ImageFormats.PAL4_RGBX5551: (_decode_rgbx5551_pixel, 4, 2, get_uint16),
         ImageFormats.PAL4_RGB888: (_decode_rgb888_pixel, 4, 3, get_uint24),
-        ImageFormats.PAL4_RGB565: (_decode_rgb565_pixel, 4, 2, get_uint16),
         ImageFormats.PAL4_RGBA8888: (_decode_rgba8888_pixel, 4, 4, get_uint32),
+        ImageFormats.PAL4_IA8: (_decode_ia8_pixel, 4, 2, get_uint16),  # N64_C4 (type 0)
+        ImageFormats.PAL4_RGB565: (_decode_rgb565_pixel, 4, 2, get_uint16),  # N64_C4 (type 1)
+        ImageFormats.PAL4_RGB5A3: (_decode_rgb5A3_pixel, 4, 2, get_uint16),  # N64_C4 (type 2)
+
         ImageFormats.PAL8_RGBX2222: (_decode_rgbx2222_pixel, 8, 1, get_uint8),
         ImageFormats.PAL8_RGBX5551: (_decode_rgbx5551_pixel, 8, 2, get_uint16),
+        ImageFormats.PAL8_BGRX5551: (_decode_bgrx5551_pixel, 8, 2, get_uint16),
         ImageFormats.PAL8_RGB888: (_decode_rgb888_pixel, 8, 3, get_uint24),
-        ImageFormats.PAL8_RGB565: (_decode_rgb565_pixel, 8, 2, get_uint16),
-        ImageFormats.PAL8_RGB5A3: (_decode_rgb5A3_pixel, 8, 2, get_uint16),
+        ImageFormats.PAL8_BGR888: (_decode_bgr888_pixel, 8, 3, get_uint24),
+        ImageFormats.PAL8_RGBX6666: (_decode_rgbx6666_pixel, 8, 3, get_uint24),
+        ImageFormats.PAL8_IA8: (_decode_ia8_pixel, 8, 2, get_uint16),  # N64_C8 (type 0)
+        ImageFormats.PAL8_RGB565: (_decode_rgb565_pixel, 8, 2, get_uint16),  # N64_C8 (type 1)
+        ImageFormats.PAL8_RGB5A3: (_decode_rgb5A3_pixel, 8, 2, get_uint16),  # N64_C8 (type 2)
         ImageFormats.PAL8_RGBA8888: (_decode_rgba8888_pixel, 8, 4, get_uint32),
         ImageFormats.PAL8_BGRA8888: (_decode_bgra8888_pixel, 8, 4, get_uint32),
+
+        ImageFormats.PAL16_IA8: (_decode_ia8_pixel, 16, 2, get_uint16),  # N64_C14X2 (type 0)
+        ImageFormats.PAL16_RGB565: (_decode_rgb565_pixel, 16, 2, get_uint16),  # N64_C14X2 (type 1)
+        ImageFormats.PAL16_RGB5A3: (_decode_rgb5A3_pixel, 16, 2, get_uint16),  # N64_C14X2 (type 2)
     }
 
     compressed_data_formats = {
@@ -265,11 +380,6 @@ class ImageDecoder:
         ImageFormats.GST822: (8, 2, 2),
     }
 
-    yuv_data_formats = {
-        # image_format: (decode_function, bpp)
-        ImageFormats.YUY2: (_decode_yuy2_pixel, 32),
-    }
-
     def _get_endianess_format(self, endianess: str) -> str:
         if endianess == "little":
             endianess_format = "<"
@@ -286,7 +396,20 @@ class ImageDecoder:
         read_offset = 0
         image_endianess_format: str = self._get_endianess_format(image_endianess)
 
-        if bits_per_pixel == 8:
+        if bits_per_pixel == 4:
+            for i in range(0, img_width * img_height, 2):
+                image_pixel: bytes = image_handler.get_bytes(read_offset, 1)
+                pixel_int: int = image_entry_read_function(image_pixel, image_endianess_format)
+                if image_endianess == "little":
+                    texture_data[i * 4:(i + 1) * 4] = decode_function(self, pixel_int & 0xf)  # noqa
+                    texture_data[(i + 1) * 4:(i + 2) * 4] = decode_function(self, (pixel_int >> 4) & 0xf)  # noqa
+                elif image_endianess == "big":
+                    texture_data[i * 4:(i + 1) * 4] = decode_function(self, (pixel_int >> 4) & 0xf)  # noqa
+                    texture_data[(i + 1) * 4:(i + 2) * 4] = decode_function(self, pixel_int & 0xf)  # noqa
+                else:
+                    raise Exception(f"Endianess not supported! Endianess: {image_endianess}")
+                read_offset += 1
+        elif bits_per_pixel == 8:
             bytes_per_pixel = 1
             for i in range(len(image_data) // bytes_per_pixel):
                 image_pixel: bytes = image_handler.get_bytes(read_offset, bytes_per_pixel)
@@ -353,23 +476,24 @@ class ImageDecoder:
             for i in range(0, img_width * img_height, 2):
                 palette_index = image_handler.get_bytes(image_offset, 1)
                 palette_index_int = struct.unpack(image_endianess_format + "B", palette_index)[0]
-                texture_data[i * 4:(i + 1) * 4] = decode_function(self, palette_data_ints[palette_index_int & 0xf])  # noqa
-                texture_data[(i + 1) * 4:(i + 2) * 4] = decode_function(self, palette_data_ints[(palette_index_int >> 4) & 0xf])  # noqa
+                if image_endianess == "little":
+                    texture_data[i * 4:(i + 1) * 4] = decode_function(self, palette_data_ints[palette_index_int & 0xf])  # noqa
+                    texture_data[(i + 1) * 4:(i + 2) * 4] = decode_function(self, palette_data_ints[(palette_index_int >> 4) & 0xf])  # noqa
+                elif image_endianess == "big":
+                    texture_data[i * 4:(i + 1) * 4] = decode_function(self, palette_data_ints[(palette_index_int >> 4) & 0xf])  # noqa
+                    texture_data[(i + 1) * 4:(i + 2) * 4] =decode_function(self, palette_data_ints[palette_index_int & 0xf])  # noqa
+                else:
+                    raise Exception(f"Endianess not supported! Endianess: {image_endianess}")
                 image_offset += 1
 
         return texture_data
 
     def _decode_compressed(self, image_data: bytes, img_width: int, img_height: int, image_format: tuple) -> bytes:
+        # TODO - replace this wrapper
         decoder_name, decoder_arg = image_format
-        pil_img = Image.frombuffer(
-            "RGBA",
-            (img_width, img_height),
-            image_data,
-            decoder_name,
-            decoder_arg,
-            "",
-        )
-        return pil_img.tobytes()
+        pillow_image: Image = self.pillow_wrapper.get_pillow_image_from_dxt_data(image_data, img_width, img_height, decoder_name, decoder_arg)
+        decoded_data: bytes = self.pillow_wrapper.get_image_data_from_pillow_image(pillow_image)
+        return decoded_data
 
     def _decode_gst(self, image_data: bytes, palette_data: bytes, img_width: int, img_height: int, image_format: tuple, convert_format: ImageFormats, is_swizzled: bool) -> bytes:
         block_width, block_height, detail_bpp = image_format
@@ -402,35 +526,6 @@ class ImageDecoder:
 
         return output_texture_data
 
-    def _decode_yuv(self, image_data: bytes, img_width: int, img_height: int, image_format: tuple):
-        decode_function, bpp = image_format
-        is_odd: bool = True if (img_width & 1) else False
-        current_yuv_offset: int = 0
-        current_pixel_number: int = 0
-        output_texture_data = bytearray(img_width * img_height * 4)
-
-        for y in range(img_height):
-            for x in range(0, img_width, 2):
-
-                Y0: float = float(image_data[current_yuv_offset])
-                U: float = float(image_data[current_yuv_offset + 1])
-                Y1: float = float(image_data[current_yuv_offset + 2])
-                V: float = float(image_data[current_yuv_offset + 3])
-
-                pixel1 = decode_function(self, Y0, U, V)
-                pixel2 = decode_function(self, Y1, U, V)
-                output_texture_data[current_pixel_number * 4:(current_pixel_number + 1) * 4] = pixel1
-                output_texture_data[(current_pixel_number + 1) * 4:(current_pixel_number + 2) * 4] = pixel2
-
-                if is_odd and x == img_width - 1:
-                    current_yuv_offset += 4
-                    current_pixel_number += 1
-                else:
-                    current_yuv_offset += 4
-                    current_pixel_number += 2
-
-        return output_texture_data
-
     def decode_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats, image_endianess: str = "little") -> bytes:
         return self._decode_generic(image_data, img_width, img_height, self.generic_data_formats[image_format], image_endianess)
 
@@ -444,4 +539,5 @@ class ImageDecoder:
         return self._decode_gst(image_data, palette_data, img_width, img_height, self.gst_data_formats[image_format], convert_format, is_swizzled)
 
     def decode_yuv_image(self, image_data: bytes, img_width: int, img_height: int, image_format: ImageFormats):
-        return self._decode_yuv(image_data, img_width, img_height, self.yuv_data_formats[image_format])
+        yuv_decoder = YUVDecoder()
+        return yuv_decoder.decode_yuv_image_main(image_data, img_width, img_height, image_format)
